@@ -14,9 +14,8 @@ from persistent.TimeStamp import TimeStamp
 import simplejson
 
 from zodbbrowser import __version__, __homepage__
-from zodbbrowser.history import ZodbObjectHistory
+from zodbbrowser.interfaces import IObjectHistory, IValueRenderer
 from zodbbrowser.state import ZodbObjectState
-from zodbbrowser.value import IValueRenderer
 from zodbbrowser.diff import compareDictsHTML
 
 
@@ -75,14 +74,15 @@ class ZodbInfoView(BrowserView):
             jar = self.jar()
             self.obj = jar.get(oid)
 
-        self.history = ZodbObjectHistory(self.obj)
+        self.history = IObjectHistory(self.obj)
         self.latest = True
         if 'tid' in self.request:
             self.state = ZodbObjectState(self.obj,
-                                         p64(int(self.request['tid'], 0)))
+                                         p64(int(self.request['tid'], 0)),
+                                         _history=self.history)
             self.latest = False
         else:
-            self.state = ZodbObjectState(self.obj)
+            self.state = ZodbObjectState(self.obj, _history=self.history)
         return self.template()
 
     def findClosestPersistent(self):
@@ -234,28 +234,42 @@ class ZodbInfoView(BrowserView):
         return [ZodbObjectAttribute(name, value, self.state.requestedTid)
                 for name, value in items]
 
+    def _loadHistoricalState(self):
+        results = []
+        for d in self.history:
+            try:
+                state = ZodbObjectState(self.obj, d['tid'],
+                                        _history=self.history).asDict()
+                error = None
+            except Exception, e:
+                state = {}
+                error = '%s: %s' % (e.__class__.__name__, e)
+            results.append(dict(state=state, error=error))
+        results.append(dict(state={}, error=None))
+        return results
+
     def listHistory(self):
         """List transactions that modified a persistent object."""
+        state = self._loadHistoricalState()
         results = []
-
-        for n, d in enumerate(self.history.history):
+        for n, d in enumerate(self.history):
             short = (str(time.strftime('%Y-%m-%d %H:%M:%S',
                                        time.localtime(d['time']))) + " "
                      + d['user_name'] + " "
                      + d['description'])
             url = self.getUrl(tid=u64(d['tid']))
-            current = d['tid'] == self.state.tid and \
-                                  self.state.requestedTid is not None
-            curState = ZodbObjectState(self.obj, d['tid']).asDict()
-            if n < len(self.history) - 1:
-                oldState = ZodbObjectState(self.obj, self.history[n + 1]['tid']).asDict()
-            else:
-                oldState = {}
+            current = (d['tid'] == self.state.tid and
+                       self.state.requestedTid is not None)
+            curState = state[n]['state']
+            oldState = state[n + 1]['state']
             diff = compareDictsHTML(curState, oldState, d['tid'])
 
             results.append(dict(short=short, utid=u64(d['tid']),
-                                href=url, current=current, diff=diff, **d))
+                                href=url, current=current,
+                                error=state[n]['error'],
+                                diff=diff, **d))
 
+        # number in reverse order
         for i in range(len(results)):
             results[i]['index'] = len(results) - i
 
